import pytest
import os

import farmOS

FARMOS_HOSTNAME = os.getenv("FARMOS_HOSTNAME")

# Credentials for Drupal RestWS Auth
FARMOS_RESTWS_USERNAME = os.getenv("FARMOS_RESTWS_USERNAME")
FARMOS_RESTWS_PASSWORD = os.getenv("FARMOS_RESTWS_PASSWORD")

# Credentials for OAuth Auth
FARMOS_OAUTH_USERNAME = os.getenv("FARMOS_OAUTH_USERNAME")
FARMOS_OAUTH_PASSWORD = os.getenv("FARMOS_OAUTH_PASSWORD")
FARMOS_OAUTH_CLIENT_ID = os.getenv("FARMOS_OAUTH_CLIENT_ID")
FARMOS_OAUTH_CLIENT_SECRET = os.getenv("FARMOS_OAUTH_CLIENT_SECRET")

valid_restws_config = \
    FARMOS_HOSTNAME is not None and FARMOS_RESTWS_USERNAME is not None and FARMOS_RESTWS_PASSWORD is not None

valid_oauth_config = \
    FARMOS_HOSTNAME is not None and FARMOS_OAUTH_USERNAME is not None and FARMOS_OAUTH_PASSWORD is not None \
    and FARMOS_OAUTH_CLIENT_ID is not None and FARMOS_OAUTH_CLIENT_SECRET is not None

farmOS_testing_server = pytest.mark.skipif(
    not (valid_restws_config or valid_oauth_config),
    reason="farmOS Testing Server not configured. Skipping farmOS test server integration tests.",
)

@pytest.fixture(scope='module')
def test_farm():
    if valid_restws_config:
        farm = farmOS.farmOS(FARMOS_HOSTNAME, FARMOS_RESTWS_USERNAME, FARMOS_RESTWS_PASSWORD)
        farm.authenticate()
        return farm

    if valid_oauth_config:
        farm = farmOS.farmOS(FARMOS_HOSTNAME, username=FARMOS_OAUTH_USERNAME, password=FARMOS_OAUTH_PASSWORD,
                             client_id=FARMOS_OAUTH_CLIENT_ID, client_secret=FARMOS_OAUTH_CLIENT_SECRET)
        farm.authenticate()
        return farm



class FarmosAPIError(Exception):
    data = {}
    code = -1
    message = "An unknown error occurred"

    def __init__(self, message=None, code=None, data=None, response=None):
        if message:
            self.message = message
        if code:
            self.code = code
        if data:
            self.data = data

    def __str__(self):
        if self.code:
            return '{}: {}'.format(self.code, self.message)
        return self.message


class NotAuthenticatedError(FarmosAPIError):
    message = "APISession not authenticated before the request was made." \
              "Call farm.authenticate() before making requests."
    pass



from configparser import ConfigParser, BasicInterpolation


class ClientConfig(ConfigParser):
    def __init__(
            self,
            profile_name=None,
            defaults=None,
            **kwargs):

        config_defaults = {
            'auto_authenticate': 'True',
            'development': 'False',
            'oauthlib_insecure_transport': 'False',
            'oauth_authorization_url': '%(hostname)s/oauth2/authorize',
            'oauth_client_id': 'farmos_api_client',
            'oauth_client_secret': '',
            'oauth_redirect_url': '%(hostname)s/api/authorized',
            'oauth_scopes': 'user_access',
            'oauth_token_url': '%(hostname)s/oauth2/token',
        }

        # Merge additional default values if provided.
        if defaults is not None:
            config_defaults = {**config_defaults, **defaults}

        # Initialize the config object.
        super().__init__(defaults=config_defaults, interpolation=BasicInterpolation())

        # Add a section for the profile.
        if profile_name is not None:
            self.add_section(profile_name)
        else:
            profile_name = 'DEFAULT'

        # Load additional kwargs into the config.
        for key, value in kwargs.items():
            if value is not None:
                self.set(profile_name, key, value)


    def write(self, path="farmos_default_config.cfg"):
        with open(path, "w") as config_file:
            super().write(config_file)



import logging

from requests import Session
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import LegacyApplicationClient

from .exceptions import NotAuthenticatedError

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

class OAuthSession(OAuth2Session):
    """OAuthSession uses OAuth2 to authenticate with farmOS

    This class stores access tokens and refresh tokens used to
    authenticate with the farmOS server for all requests.

    Keyword Arguments:
        hostname - the farmOS hostname (without protocol)
        client_id - the farmOS API Client ID
        client_secret - the farmOS API Client Secret
        username - the farmOS username (for OAuth2 Password Grant)
        password - the farmOS user's password (for OAuth2 Password Grant)
    """

    def __init__(self, hostname,
                 grant_type,
                 client_id,
                 client_secret=None,
                 scope=None,
                 username=None,
                 password=None,
                 token=None,
                 redirect_uri=None,
                 token_url=None,
                 authorization_url=None,
                 token_updater=None,
                 *args, **kwargs):

        # Initialize the session as not authenticated.
        self.authenticated = False
        self.csrf_token = None

        # Create a dictionary of credentials required to pass along with Refresh Tokens
        # Required to generate a new access token
        auto_refresh_kwargs = {'client_id': client_id,
                               'client_secret': client_secret
                               }

        # Validate the grant type.
        valid_grant_types = ["Authorization", "Password"]
        if grant_type not in valid_grant_types:
            raise Exception(grant_type + " is not a supported OAuth Grant Type")

        # Save the Grant Type
        self.grant_type = grant_type

        if grant_type == "Authorization":
            super(OAuthSession, self).__init__(token=token,
                                               client_id=client_id,
                                               redirect_uri=redirect_uri,
                                               auto_refresh_url=token_url,
                                               auto_refresh_kwargs=auto_refresh_kwargs,
                                               token_updater=token_updater,
                                               scope=scope)
        elif grant_type == "Password":
            super(OAuthSession, self).__init__(token=token,
                                               client=LegacyApplicationClient(client_id=client_id),
                                               auto_refresh_url=token_url,
                                               auto_refresh_kwargs=auto_refresh_kwargs,
                                               token_updater=token_updater,
                                               scope=scope)

        # Check if 'user_access' scope is included.
        # If so, indicate that we have full User Access and should request
        # a CSRF token to allow making non-GET requests.
        self.has_user_access = scope.find('user_access') > -1

        # Save values for later use.
        self._token_url = token_url
        self._authorization_base_url = authorization_url
        self._client_id = client_id
        self._client_secret = client_secret
        self._username = username
        self._password = password
        self.hostname = hostname

    def authenticate(self):
        """Authenticates with the farmOS site.

        Returns True or False indicating whether or not
        the authentication was successful
        """
        token = self.token

        if 'access_token' not in token:
            logger.debug('Retrieving new OAuth Token.')

            if self.grant_type == "Authorization":
                authorization_url, state = self.authorization_url(self._authorization_base_url,
                                                                  access_type="offline",
                                                                  prompt="select_account")

                print('Please go here and authorize,', authorization_url)

                # Get the authorization verifier code from the callback url
                redirect_response = input('Paste the full redirect URL here:')

                # Fetch the access token
                token = self.fetch_token(self._token_url,
                                         client_secret=self._client_secret,
                                         authorization_response=redirect_response)

            elif self.grant_type == "Password":
                token = self.fetch_token(token_url=self._token_url,
                                         client_id=self._client_id,
                                         client_secret=self._client_secret,
                                         username=self._username,
                                         password=self._password,
                                         scope=self.scope)

            logger.debug('Fetched OAuth Access Token %s', token)

            # Save the token.
            logger.debug('Saving token with token_updater utility.')
            self.token_updater(token)

        # If the OAuth Token has user access, request a session token
        # from the RESTful Web Services module. This is required for
        # non HTTP GET requests.
        if self.has_user_access:
            self.csrf_token = _get_csrf_token(self)

        return self.is_authenticated(check=True)

    def http_request(self, path, method='GET', options=None, params=None, force=False):
        """Raw HTTP request helper function.

        Keyword arguments:
        :param path: the URL path.
        :param method: the HTTP method.
        :param options: a dictionary of data and parameters to pass on to the request.
        :param params: URL query parameters.
        :param force: Force the request regardless of Authenticated status.
        :return: requests response object.
        """

        # If the session has not been authenticated
        # and the request does not have force=True,
        # raise NotAuthenticatedError
        if not self.is_authenticated() and not force:
            raise NotAuthenticatedError()

        headers = {}
        # Automatically add the token to the request, if it exists.
        # Give precedence to a token passed in with options['headers']['X-CSRF-Token'].
        if options and 'headers' in options and 'X-CSRF-Token' in options['headers']:
            headers['X-CSRF-Token'] = options['headers']['X-CSRF-Token']
        if self.csrf_token is not None:
            headers['X-CSRF-Token'] = self.csrf_token

        # Return response from the _http_request helper function.
        return _http_request(self, path, method, options, params, headers)

    def is_authenticated(self, check=False):
        """Helper function that returns True or False if the Session is Authenticated"""
        # If the OAuth Token has user access, check if the session
        # has a csrf_token. This is required for non HTTP GET requests.
        if self.has_user_access and self.csrf_token is None:
            return False

        if check is False:
            return self.authenticated
        else:
            return _is_authenticated(self)


# Use a Requests Session to store cookies across requests.
#   http://docs.python-requests.org/en/master/user/advanced/#session-objects
class DrupalAuthSession(Session):

    """APISession handles all HTTP requests for the farmOS API

    This class stores cookies and tokens required for the Drupal
    RESTFul Web Services HTTP Session Authentication method.
    These values are use in all requests made to the farmOS host.

    Keyword Arguments:
    hostname - the farmOS hostname (without protocol)
    username - the farmOS username
    password - the farmOS user's password
    """

    def __init__(self, hostname, username, password, *args, **kwargs):
        super(DrupalAuthSession, self).__init__()

        # Initialize the session as not authenticated.
        self.authenticated = False

        # Store farmOS authentication credentials
        self.hostname = hostname
        self.username = username
        self.password = password

        # Store the authentication token
        self.csrf_token = None

    def authenticate(self):
        """Authenticates with the farmOS site.

        Returns True or False indicating whether or not
        the authentication was successful
        """
        logger.debug('Authenticating with Drupal Login Form')
        # Prepare the data payload
        options = {
            'data': {
                'name': self.username,
                'pass': self.password,
                'form_id': 'user_login',
            }
        }

        # Login with the username and password to get a cookie.
        response = self.http_request('user/login', 'POST', options, force=True)
        if response:
            if response.status_code != 200:
                return False

        # Request a session token from the RESTful Web Services module
        self.csrf_token = _get_csrf_token(self)

        return self.is_authenticated(check=True)

    def http_request(self, path, method='GET', options=None, params=None, force=False):
        """Make an HTTP request.

        Keyword arguments:
        :param path: the URL path.
        :param method: the HTTP method.
        :param options: a dictionary of data and parameters to pass on to the request.
        :param params: URL query parameters.
        :param force: Force the request regardless of Authenticated status.
        :return: requests response object.
        """

        # If the session has not been authenticated
        # and the request does not have force=True,
        # raise NotAuthenticatedError
        if not self.is_authenticated() and not force:
            raise NotAuthenticatedError()

        # Start a headers dictionary.
        headers = {}

        # Automatically add the token to the request, if it exists.
        # Give precedence to a token passed in with options['headers']['X-CSRF-Token'].
        if options and 'headers' in options and 'X-CSRF-Token' in options['headers']:
            headers['X-CSRF-Token'] = options['headers']['X-CSRF-Token']
        if self.csrf_token:
            headers['X-CSRF-Token'] = self.csrf_token

        # Return response from the _http_request helper function.
        return _http_request(self, path, method, options, params, headers)

    def is_authenticated(self, check=False):
        """Helper function that returns True or False if the Session is Authenticated"""
        # Check if the session has a csrf_token
        if self.csrf_token is None:
            return False

        if check is False:
            return self.authenticated
        else:
            return _is_authenticated(self)


def _http_request(session, path, method='GET', options=None, params=None, headers=None):
    """Raw HTTP request helper function.

    Keyword arguments:
    :param session: The requests Session object to call a request with.
    :param path: the URL path.
    :param method: the HTTP method.
    :param options: a dictionary of data and parameters to pass on to the request.
    :param params: URL query parameters.
    :param headers: Dictionary of HTTP headers to include in the request.
    :return: requests response object.

    """
    # Strip protocol, hostname, leading/trailing slashes, and whitespace from the path.
    path = path.strip('/')
    path = path.strip()

    # Assemble the URL.
    url = '{}/{}'.format(session.hostname, path)

    # Automatically follow redirects, unless this is a POST request.
    # The Python requests library converts POST to GET during a redirect.
    # Allow this to be overridden in options.
    allow_redirects = True
    if method in ['POST', 'PUT']:
        allow_redirects = False
    if options and 'allow_redirects' in options:
        allow_redirects = options['allow_redirects']

    # If there is data to be sent, include it.
    data = None
    if options and 'data' in options:
        data = options['data']

    # If there is a json data to be sent, include it.
    json = None
    if options and 'json' in options:
        json = options['json']

    if headers is None:
        headers = {}

    # Perform the request.
    response = session.request(method,
                               url,
                               headers=headers,
                               allow_redirects=allow_redirects,
                               data=data,
                               json=json,
                               params=params)

    # If this is a POST request, and a redirect occurred, attempt to re-POST.
    redirect_codes = [300, 301, 302, 303, 304, 305, 306, 307, 308]
    if method in ['POST', 'PUT'] and response.status_code in redirect_codes:
        if response.headers['Location']:
            response = session.request(method,
                                       response.headers['Location'],
                                       headers=headers,
                                       allow_redirects=True,
                                       data=data,
                                       json=json, params=params)

    # Return the response.
    return response

def _get_csrf_token(session):
    """Helper function to retrieve a CSRF Session Token"""
    csrf_token = None

    logger.debug('Retrieving new csrf_token.')

    # Request a session token from the RESTful Web Services module
    response = session.http_request('restws/session/token', force=True)
    if response:
        if response.status_code == 200:
            csrf_token = response.text

    return csrf_token


def _is_authenticated(session):
    """Helper function to check if the Session is authenticated."""

    try:
        response = session.http_request(path='farm.json', force=True)
    except NotAuthenticatedError:
        return False

    if response.status_code == 200:
        session.authenticated = True
        return True
    else:
        session.authenticated = False
        return False



import setuptools

setup_requires=[
    'pytest-runner',
]

install_requires=[
    'requests',
    'requests-oauthlib',
    ]

tests_require=[
    'pytest',
],

with open('README.md', 'r') as fh:
    readme = fh.read()

setuptools.setup(
    name='farmOS',
    version='0.1.6b1',
    author='farmOS team',
    author_email='mike@mstenta.net',
    description='A Python library for interacting with farmOS over API. ',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/farmOS/farmOS.py',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3',
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    test_suite='pytest',
)



import logging
from datetime import datetime
from urllib.parse import urlparse, urlunparse

from .session import DrupalAuthSession, OAuthSession
from .client import LogAPI, AssetAPI, TermAPI, AreaAPI
from .config import ClientConfig

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

class farmOS:

    """Create a new farmOS instance.

    Keyword Arguments:
    hostname - the farmOS hostname (without protocol)
    username - the farmOS username
    password - the farmOS user's password

    Attributes:
    session - an APISession object that handles HTTP requests

    """

    def __init__(self,
                 hostname=None,
                 username=None,
                 password=None,
                 client_id=None,
                 client_secret=None,
                 scope=None,
                 config=None,
                 config_file=None,
                 profile_name=None,
                 token_updater=None):

        logger.debug('Creating farmOS client.')

        # Start a list of config files.
        config_file_list = ['farmos_default_config.cfg']

        # Append additional config files.
        self.config_file = None
        if config_file is not None:
            if isinstance(config_file, str):
                logger.debug('Using config file: %s', config_file)
                config_file_list.append(config_file)
                self.config_file = config_file
            else:
                raise Exception("Config file must be a string.")

        # Check for a provided configuration object.
        if isinstance(config, ClientConfig):
            logger.debug('Using provided ClientConfig object.')
            self.config = config
        elif config is not None and not isinstance(config, ClientConfig):
            raise Exception("Config is not a ClientConfig object.")
        # Create a new object if none is provided.
        elif config is None:
            logger.debug('No ClientConfig object provided, using defaults.')
            self.config = ClientConfig(profile_name=profile_name)

        # Read config files.
        self.config.read(config_file_list)
        logger.debug('Loaded config files: %s', config_file_list)

        # Use a profile if provided.
        self.profile = None
        self.profile_name = "DEFAULT"
        if profile_name is not None:
            self.use_profile(profile_name, create_profile=True)
            logger.debug('Using profile name "%s"', profile_name)

        # Load the config boolean for development mode.
        self.development = self.config.getboolean(self.profile_name, "development", fallback=False)
        if self.development:
            logger.warning('Development mode enabled.')

        # Allow authentication over HTTP in development mode
        # or if the oauth_insecure_transport config is enabled.
        oauth_insecure_transport = self.config.getboolean(self.profile_name,
                                                          "oauthlib_insecure_transport",
                                                          fallback=False)
        if oauth_insecure_transport:
            logger.warning('OAuth Insecure Transport enabled in configuration.')

        if self.development or oauth_insecure_transport:
            import os
            os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'
            logger.warning('OAuth Insecure Transport is enabled.')

        # Load the token updater function.
        # Default to simple token_saver to save tokens to config.
        self.token_updater = self.save_token
        if token_updater is not None:
            logger.debug('Using provided token_updater utility.')
            self.token_updater = token_updater


        self.session = None

        # A username or client_id is required for authentication to farmOS.
        if username is None and client_id is None:
            # Try to load values from config profile.
            logger.debug('No username or client_id provided. Loading from config...')

            hostname = self.config.get(self.profile_name, "hostname", fallback=None)
            username = self.config.get(self.profile_name, "username", fallback=None)
            password = self.config.get(self.profile_name, "password", fallback=None, raw=True)
            client_id = self.config.get(self.profile_name, "oauth_client_id", fallback=None)
            client_secret = self.config.get(self.profile_name, "oauth_client_secret", fallback=None)
            scope = self.config.get(self.profile_name, "oauth_scope", fallback=None)

            # Check if required values were populated.
            if username is None and client_id is None:
                raise Exception("No authentication method provided.")

        if hostname is not None:
            valid_schemes = ["http", "https"]
            default_scheme = "http" if self.development else "https"
            parsed_url = urlparse(hostname)

            # Validate the hostname.
            # Add a default scheme if not provided.
            if not parsed_url.scheme:
                parsed_url = parsed_url._replace(scheme=default_scheme)
                logger.debug('No scheme provided. Using %s', default_scheme)

            # Check for a valid scheme.
            if parsed_url.scheme not in valid_schemes:
                raise Exception("Not a valid scheme.")

            # If not netloc was provided, it was probably parsed as the path.
            if not parsed_url.netloc and parsed_url.path:
                parsed_url = parsed_url._replace(netloc=parsed_url.path)
                parsed_url = parsed_url._replace(path='')

            # Check for netloc.
            if not parsed_url.netloc:
                raise Exception("Invalid hostname. Must have netloc.")

            # Don't allow path, params, or query.
            if parsed_url.path or parsed_url.params or parsed_url.query:
                raise Exception("Hostname cannot include path and query parameters.")

            # Build the url again to include changes.
            hostname = urlunparse(parsed_url)
            logger.debug('Complete hostname configured as %s', hostname)

            # Save the hostname in the config.
            self.config.set(self.profile_name, "hostname", hostname)

        else:
            raise Exception("No hostname provided and could not be loaded from config.")

        # Check if we have a token
        has_token = False
        if self.profile and 'access_token' in dict(self.profile):
            has_token = True

        # Ask for password if username is given without a password
        # and the profile does not have an OAuth Token.
        if username is not None and password is None and not has_token:
            from getpass import getpass
            password = getpass('Enter password: ')

        # If a client_id is supplied, try to create an OAuth Session
        if client_id is not None:
            logger.debug('Creating an OAuth Session...')
            token_url = self.config.get(self.profile_name, "oauth_token_url")

            # Load saved Authentication Profile from config.
            token = None
            if self.has_profile():
                logger.debug('Loading Authentication Profile from config.')

                # Save OAuth Client ID to config.
                if client_id is not None:
                    self.config.set(self.profile_name, "oauth_client_id", client_id)
                if client_secret is not None:
                    self.config.set(self.profile_name, "oauth_client_secret", client_secret)
                if scope is not None:
                    self.config.set(self.profile_name, "oauth_scope", scope)

                # Initialize an empty token dict.
                token = {}

                if 'access_token' in self.config[profile_name]:
                    token['access_token'] = self.config[profile_name]['access_token'],

                if 'refresh_token' in self.config[profile_name]:
                    token['refresh_token'] = self.config[profile_name]['refresh_token'],

                # Check the token expiration time.
                if 'expires_at' in self.config[profile_name]:
                    # Create datetime objects for comparison.
                    now = datetime.now()
                    expiration_time = datetime.fromtimestamp(float(self.config[profile_name]['expires_at']))

                    # Calculate seconds until expiration.
                    timedelta = expiration_time - now
                    expires_in = timedelta.total_seconds()

                    # Update the token expires_in value
                    token['expires_in'] = expires_in

            # Create an OAuth Session with the Password Credentials Grant.
            if username is not None and password is not None:
                logger.debug('Using OAuth Password Credentials Grant.')

                self.config.set(self.profile_name, "username", username)
                self.session = OAuthSession(grant_type="Password",
                                            hostname=hostname,
                                            client_id=client_id,
                                            client_secret=client_secret,
                                            scope=scope,
                                            username=username,
                                            password=password,
                                            token=token,
                                            token_url=token_url,
                                            token_updater=self.token_updater)

            # Create an OAuth Session with the Authorization Code Grant.
            else:
                logger.debug('Starting OAuth Authorization Code flow.')

                # Load saved OAuth URLs from config.
                authorization_url = self.config.get(self.profile_name, "oauth_authorization_url")
                redirect_url = self.config.get(self.profile_name, "oauth_redirect_url")

                # Create an OAuth Session
                self.session = OAuthSession(grant_type="Authorization",
                                            hostname=hostname,
                                            client_id=client_id,
                                            client_secret=client_secret,
                                            scope=scope,
                                            token=token,
                                            redirect_uri=redirect_url,
                                            token_url=token_url,
                                            authorization_url=authorization_url,
                                            token_updater=self.token_updater)

        # Fallback to DrupalAPISession
        if client_id is None and username is not None and password is not None:
            logger.debug('Authenticating with Drupal RESTws Authentication.')
            # Create a session with requests
            self.session = DrupalAuthSession(hostname, username, password)

        self._username = username
        self._client_id = client_id
        self._client_secret = client_secret

        if self.session is None:
            raise Exception("Could not create a session object. Supply authentication credentials when "
                            "initializing a farmOS Client.")

        self.log = LogAPI(self.session)
        self.asset = AssetAPI(self.session)
        self.area = AreaAPI(self.session)
        self.term = TermAPI(self.session)

        # Authenticate the session if auto_authenticate is enabled
        # and the session has not already been authenticated.
        # (OAuthSession will authenticate if an existing token was provided at startup)
        auto_authenticate = self.config.getboolean(self.profile_name, "auto_authenticate", fallback=True)
        if auto_authenticate and not self.session.is_authenticated():
            logger.debug('Auto authenticate set to true.')
            status = self.session.authenticate()
            logger.debug('Authenticated: %s', status)

    def authenticate(self):
        """Authenticates with the farmOS site.

        Returns True or False indicating whether or not
        the authentication was successful
        """
        logger.debug('Authenticating with farmOS server.')
        return self.session.authenticate()

    def info(self, path='farm.json'):
        """Retrieve info about the farmOS instance"""
        logger.debug('Retrieving farmOS server info.')
        response = self.session.http_request(path)
        if response.status_code == 200:
            return response.json()

        return []

    def save_token(self, token):
        """Save an OAuth Token to config for later use.

        This method accepts an OAuth token and saves values to the Authentication
        section of the farm.config ClientConfig object. It is primarily used as
        a callback for the requests-oauthlib OAuth2Session automatic token refreshing
        functionality. But this method could be used by others to supply a farmOS client
        with existing OAuth tokens to use (and persist).

        :param token: OAuth token dict.
        :return: None.
        """
        # Only save values if a profile name was defined.
        if self.has_profile():
            profile_name = self.get_profile_name()
            logger.debug('Saving new OAuth token to profile %s', profile_name)

            if 'access_token' in token:
                self.config.set(profile_name, "access_token", token['access_token'])

            if 'expires_in' in token:
                # token['expires_in'] is an int, the access_token lifetime.
                # Must be saved as a string in the config.
                self.config.set(profile_name, "expires_in", str(token['expires_in']))

            if 'token_type' in token:
                self.config.set(profile_name, "token_type", token['token_type'])

            if 'refresh_token' in token:
                self.config.set(profile_name, "refresh_token", token['refresh_token'])

            if 'expires_at' in token:
                # token['expires_at'] is a float, the access_token expiration time.
                # requests-oauthlib generates this value as
                #       expires_at = time.time() + expires_in
                # Must be saved as a string in the config.
                self.config.set(profile_name, "expires_at", str(token['expires_at']))

        if self.config_file is None:
            logger.debug('No profile configured. New OAuth token will not be saved to config.')
        else:
            self.config.write(path=self.config_file)

    def create_profile(self, profile_name):
        """Creates a Section for profile_name in farm.config."""
        if not self._profile_exists(profile_name):
            self.config.add_section(profile_name)
            return True
        else:
            # TODO: Write test for duplicate profile names.
            raise Exception("Profile '" + profile_name + "' already exists.")

    def has_profile(self, profile_name=None):
        """Returns True or False if the client is configured with a profile.

        Also returns whether a profile_name is found in the config.
        """
        if profile_name is not None:
            return self._profile_exists(profile_name)
        else:
            return self.profile is not None

    def get_profile_name(self):
        """Returns the current profile name."""
        if self.has_profile():
            return self.profile_name
        else:
            raise Exception("No profile being used.")

    def use_profile(self, profile_name, create_profile=False):
        """Set the authentication profile to use from farm.config."""
        profile = self._get_profile_config(profile_name)

        if profile is None:
            if create_profile is True:
                self.create_profile(profile_name)
                profile = self._get_profile_config(profile_name)
            else:
                # TODO: Write test for no profile name.
                raise Exception("Profile '" + profile_name + "' does not exist.")

        self.profile = profile
        self.profile_name = profile_name

    def _get_profile_config(self, profile_name=None):
        """Helper function that returns the current profile config, or the config or profile_name."""
        if self._profile_exists(profile_name):
            return self.config[profile_name]
        else:
            return None

    def _profile_exists(self, profile_name):
        """Helper function to check if a profile for profile_name exists."""
        if isinstance(profile_name, str):
            return self.config.has_section(profile_name)
        else:
            # TODO: Write test for invalid profile_name.
            raise Exception("profile_name not a String.")
